"""Library for polling dataplanes for statistics."""

# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import random
import time

from os_ken.lib import hub
from os_ken.ofproto import ofproto_v1_3 as ofp
from os_ken.ofproto import ofproto_v1_3_parser as parser

from faucet.valve_of import devid_present
from faucet.valve_of_old import OLD_MATCH_FIELDS


class GaugePoller:
    """Abstraction for a poller for statistics."""

    def __init__(self, conf, logname, prom_client):
        self.dp = conf.dp  # pylint: disable=invalid-name
        self.conf = conf
        self.prom_client = prom_client
        self.reply_pending = False
        self.ryudp = None
        self.logger = logging.getLogger(logname + ".{0}".format(self.conf.type))
        # _running indicates that the watcher is receiving data
        self._running = False
        self.req = None

    def report_dp_status(self, dp_status):
        """Report DP status."""
        self.prom_client.dp_status.labels(
            **dict(dp_id=hex(self.dp.dp_id), dp_name=self.dp.name)
        ).set(
            dp_status
        )  # pylint: disable=no-member

    def start(self, ryudp, active):
        """Start the poller."""
        self.ryudp = ryudp
        self._running = True
        if active:
            self.logger.info("starting")

    def stop(self):
        """Stop the poller."""
        self.logger.info("stopping")
        self._running = False

    def running(self):
        """Return True if the poller is running."""
        return self._running

    @staticmethod
    def is_active():
        """Return True if the poller is controlling the request loop for its stat"""
        return False

    def send_req(self):
        """Send a stats request to a datapath."""
        raise NotImplementedError  # pragma: no cover

    def no_response(self):
        """Called when a polling cycle passes without receiving a response."""

        dpid_str = ""
        if self.req and hasattr(self.req, "datapath"):
            dpid_str = "DPID %s (%s)" % (
                self.req.datapath.id,
                hex(self.req.datapath.id),
            )

        self.logger.info("%s no response to %s", dpid_str, self.req)

    def update(self, rcv_time, msg):
        """Handle the responses to requests.

        Called when a reply to a stats request sent by this object is received
        by the controller.

        It should acknowledge the receipt by setting self.reply_pending to
        false.

        Args:
            rcv_time: the time the response was received
            msg: the stats reply message
        """
        # TODO: it may be worth while verifying this is the correct stats
        # response before doing this
        if not self.running():
            self.logger.debug("update received when not running")
            return
        self.reply_pending = False
        self._update(rcv_time, msg)

    def _format_stat_pairs(self, _delim, _stat):
        return ()

    def _dp_stat_name(self, _stat, _stat_name):
        return ""

    def _rcv_time(self, rcv_time):
        return time.strftime("%b %d %H:%M:%S", time.localtime(rcv_time))

    def _update(self, rcv_time, msg):
        if not self.conf.file:
            return
        rcv_time_str = self._rcv_time(rcv_time)
        log_lines = []
        for stat in msg.body:
            for stat_name, stat_val in self._format_stat_pairs("-", stat):
                dp_stat_name = self._dp_stat_name(stat, stat_name)
                log_lines.append(
                    self._update_line(rcv_time_str, dp_stat_name, stat_val)
                )
        with open(self.conf.file, "a", encoding="utf-8") as logfile:
            logfile.writelines(log_lines)

    @staticmethod
    def _format_stats(delim, stat_pairs):
        formatted_stats = []
        for stat_name_list, stat_val in stat_pairs:
            stat_name = delim.join(stat_name_list)
            # OVS reports unsupported statistics as all-1-bits (UINT64_MAX)
            if stat_val == 2**64 - 1:
                stat_val = 0
            formatted_stats.append((stat_name, stat_val))
        return formatted_stats

    @staticmethod
    def _update_line(rcv_time_str, stat_name, stat_val):
        return "\t".join((rcv_time_str, stat_name, str(stat_val))) + "\n"


class GaugeThreadPoller(GaugePoller):
    """A ryu thread object for sending and receiving OpenFlow stats requests.

    The thread runs in a loop sending a request, sleeping then checking a
    response was received before sending another request.

    The methods send_req, update and no_response should be implemented by
    subclasses.
    """

    def __init__(self, conf, logname, prom_client):
        super().__init__(conf, logname, prom_client)
        self.thread = None
        self.interval = self.conf.interval
        self.ryudp = None

    def start(self, ryudp, active):
        self.stop()
        super().start(ryudp, active)
        if active:
            self.thread = hub.spawn(self)
            self.thread.name = "GaugeThreadPoller"

    def stop(self):
        super().stop()
        if self.is_active():
            hub.kill(self.thread)
            hub.joinall([self.thread])
            self.thread = None

    def is_active(self):
        return self.thread is not None

    def __call__(self):
        """Send request loop.

        Delays the initial request for a random interval to reduce load.
        Then sends a request to the datapath, waits the specified interval and
        checks that a response has been received in a loop."""
        # TODO: this should use a deterministic method instead of random
        hub.sleep(random.randint(1, self.conf.interval))
        while True:
            self.send_req()
            self.reply_pending = True
            hub.sleep(self.conf.interval)
            if self.reply_pending:
                self.no_response()

    def send_req(self):
        """Send a stats request to a datapath."""
        raise NotImplementedError  # pragma: no cover


class GaugeMeterStatsPoller(GaugeThreadPoller):
    """Poll for all meter stats."""

    def send_req(self):
        if self.ryudp:
            self.req = parser.OFPMeterStatsRequest(self.ryudp, 0, ofp.OFPM_ALL)
            self.ryudp.send_msg(self.req)


class GaugePortStatsPoller(GaugeThreadPoller):
    """Periodically sends a port stats request to the datapath and parses
    and outputs the response.
    """

    def send_req(self):
        if self.ryudp:
            self.req = parser.OFPPortStatsRequest(self.ryudp, 0, ofp.OFPP_ANY)
            self.ryudp.send_msg(self.req)

    def _format_stat_pairs(self, delim, stat):
        stat_pairs = (
            (("packets", "out"), stat.tx_packets),
            (("packets", "in"), stat.rx_packets),
            (("bytes", "out"), stat.tx_bytes),
            (("bytes", "in"), stat.rx_bytes),
            (("dropped", "out"), stat.tx_dropped),
            (("dropped", "in"), stat.rx_dropped),
            (("errors", "out"), stat.tx_errors),
            (("errors", "in"), stat.rx_errors),
        )
        return self._format_stats(delim, stat_pairs)


class GaugeFlowTablePoller(GaugeThreadPoller):
    """Periodically dumps the current datapath flow table as a yaml object.

    Includes a timestamp and a reference ($DATAPATHNAME-flowtables). The
    flow table is dumped as an OFFlowStatsReply message (in yaml format) that
    matches all flows.
    """

    def send_req(self):
        if self.ryudp:
            match = parser.OFPMatch()
            self.req = parser.OFPFlowStatsRequest(
                self.ryudp, 0, ofp.OFPTT_ALL, ofp.OFPP_ANY, ofp.OFPG_ANY, 0, 0, match
            )
            self.ryudp.send_msg(self.req)

    def _parse_flow_stats(self, stats):
        """Parse flow stats reply message into tags/labels and byte/packet counts."""
        packet_count = int(stats["packet_count"])
        byte_count = int(stats["byte_count"])
        instructions = stats["instructions"]
        tags = {
            "dp_name": self.dp.name,
            "dp_id": hex(self.dp.dp_id),
            "table_id": int(stats["table_id"]),
            "priority": int(stats["priority"]),
            "inst_count": len(instructions),
            "cookie": int(stats["cookie"]),
        }
        oxm_matches = stats["match"]["OFPMatch"]["oxm_fields"]
        for oxm_match in oxm_matches:
            oxm_tlv = oxm_match["OXMTlv"]
            mask = oxm_tlv["mask"]
            val = oxm_tlv["value"]
            orig_field = oxm_tlv["field"]
            if mask is not None:
                val = "/".join((str(val), str(mask)))
            field = OLD_MATCH_FIELDS.get(orig_field, orig_field)
            tags[field] = val
            if field == "vlan_vid" and mask is None:
                tags["vlan"] = devid_present(int(val))
        return (
            ("flow_packet_count", tags, packet_count),
            ("flow_byte_count", tags, byte_count),
        )


class GaugePortStatePoller(GaugePoller):
    """Abstraction for port state poller."""

    def send_req(self):
        """Send a stats request to a datapath."""
        raise NotImplementedError  # pragma: no cover

    def no_response(self):
        """Called when a polling cycle passes without receiving a response."""
        raise NotImplementedError  # pragma: no cover



"""Manage Fake OF tables for unit tests"""

# pylint: disable=too-many-lines

# Copyright (C) 2015 Research and Innovation Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import argparse
import json
import ast
import heapq
import pprint

from collections import OrderedDict

from bitstring import Bits

from os_ken.ofproto import ofproto_v1_3 as ofp
from os_ken.ofproto import ofproto_v1_3_parser as parser
from os_ken.ofproto import ofproto_parser as ofp_parser
from os_ken.lib import addrconv


CONTROLLER_PORT = 4294967293
IN_PORT = 4294967288


class FakeOFTableException(Exception):
    """Indicates an erroneous flow or group mod"""


class DFS:
    """Provides a way of tracking the search through the FakeOFNetwork"""

    visited = None
    heap = None

    def __init__(self):
        self.visited = {}
        self.heap = []

    def visit(self, dp_id, pkt):
        """
        Notifies the DFS that a packet has visited the dp_id
        Args:
            dp_id: The DP ID for the node that is being visited
            pkt: The packet that is visiting the node
        """
        self.visited.setdefault(dp_id, [])
        if pkt not in self.visited[dp_id]:
            self.visited[dp_id].append(pkt)

    def has_visited(self, dp_id, pkt):
        """
        Returns true if the packet has visited the node DP ID before
        Args:
            dp_id: The DP ID for the node is being visited
            pkt: The packet that is visiting the node
        """
        if dp_id in self.visited:
            if pkt in self.visited[dp_id]:
                return True
        return False

    def peek(self):
        """
        Returns the first item in the heap (with the highest priority (smallest value))
        with popping from the heap
        Returns:
            dp_id, pkt
        """
        if not self.heap:
            return None, None
        item = self.heap[0]
        return item[1][0], item[1][1]

    def push(self, dp_id, pkt, priority):
        """
        Pushes the dp_id and pkt onto the heap with priority
        Args:
            dp_id:
            pkt:
            priority:
        """
        heapq.heappush(self.heap, (priority, (dp_id, tuple(pkt.items()))))

    def pop(self):
        """
        Obtains the item with the highest priority
        Returns:
            dp_id, pkt
        """
        if not self.heap:
            return None, None
        item = heapq.heappop(self.heap)
        return item[1][0], item[1][1]


class FakeOFNetwork:
    """
    FakeOFNetwork is a virtual openflow pipeline used for testing openflow controllers
    The network contains multiple FakeOFTables to represent multiple switches in a network
    """

    def __init__(self, valves_manager, num_tables, requires_tfm=True):
        """
        Args:
            valves_manager (ValvesManager): Valves manager class to resolve stack traversals
            num_tables (int): The number of tables to configure in each FakeOFTable
            requires_tfm (bool): Whether TFMs are required
        """
        self.valves_manager = valves_manager
        self.tables = {}
        for dp_id in self.valves_manager.valves:
            self.tables[dp_id] = FakeOFTable(dp_id, num_tables, requires_tfm)

    def apply_ofmsgs(self, dp_id, ofmsgs, ignore_errors=False):
        """Applies ofmsgs to a FakeOFTable for DP ID"""
        self.tables[dp_id].apply_ofmsgs(ofmsgs, ignore_errors=ignore_errors)

    def print_table(self, dp_id):
        """Prints the table in string format to STDERR"""
        sys.stderr.write("TABLE %x" % dp_id)
        sys.stderr.write(str(self.tables[dp_id]) + "\n")
        sys.stderr.write("======================\n\n")

    def shortest_path_len(self, src_dpid, dst_dpid):
        """Returns the length of the shortest path from the source to the destination"""
        if src_dpid == dst_dpid:
            return 1
        src_valve = self.valves_manager.valves[src_dpid]
        dst_valve = self.valves_manager.valves[dst_dpid]
        if src_valve.dp.stack and dst_valve.dp.stack:
            return len(src_valve.dp.stack.shortest_path(dst_valve.dp.name))
        return 2

    def is_output(self, match, src_dpid, dst_dpid, port=None, vid=None, trace=False):
        """
        Traverses a packet through the network until we have searched everything
        or successfully output a packet to the destination with expected port and vid
        If port is None return True if output to any port (including special ports)
        regardless of VLAN tag.
        If vid is None return True if output to specified port regardless of VLAN tag.
        If vid OFPVID_PRESENT bit is 0, return True if output packet does not have
        a VLAN tag OR packet OFPVID_PRESENT is 0
        Args:
            match (dict): A dictionary keyed by header field names with values
            src_dpid: The source DP ID of the match packet entering the Fake OF network
            dst_dpid: The expected destination DP ID of the packet match
            port: The expected output port on the destination DP
            vid: The expected output vid on the destination DP
            trace (bool): Print the trace of traversing the tables
        Returns:
            true if packets with match fields is output to port with correct VLAN
        """
        found = False
        dfs = DFS()
        priority = self.shortest_path_len(src_dpid, dst_dpid)
        pkt = match.copy()
        dfs.push(src_dpid, pkt, priority)
        dfs.visit(src_dpid, pkt)
        while not found:
            # Search through the packet paths until we have searched everything or
            #   successfully output the packet to the destination in the expected format
            dp_id, pkt = dfs.pop()
            if dp_id is None or pkt is None:
                break
            pkt = dict(pkt)
            if dp_id == dst_dpid:
                # A packet has reached the destination, so test for the output
                found = self.tables[dp_id].is_full_output(pkt, port, vid, trace=trace)
                if not found and trace:
                    # A packet on the destination DP is not output in the expected state so
                    #   continue searching (flood reflection)
                    sys.stderr.write("Output is away from destination\n")
            if not found:
                # Packet not reached destination, so continue traversing
                if trace:
                    sys.stderr.write("FakeOFTable %s: %s\n" % (dp_id, pkt))
                port_outputs = self.tables[dp_id].get_port_outputs(pkt, trace=trace)
                valve = self.valves_manager.valves[dp_id]
                for out_port, out_pkts in port_outputs.items():
                    if out_port == IN_PORT:
                        # Rebind output to the packet in_port value
                        out_port = pkt["in_port"]
                    if out_port not in valve.dp.ports:
                        # Ignore output to improper ports & controller
                        # TODO: Here we should actually send the packet to the
                        #   controller, and maybe install necessary rules to
                        #   help testing routing implementations
                        continue
                    for out_pkt in out_pkts:
                        port_obj = valve.dp.ports[out_port]
                        if port_obj.stack:
                            # Need to continue traversing through the FakeOFNetwork
                            adj_port = port_obj.stack["port"]
                            adj_dpid = port_obj.stack["dp"].dp_id
                            new_pkt = out_pkt.copy()
                            new_pkt["in_port"] = adj_port.number
                            if not dfs.has_visited(adj_dpid, new_pkt):
                                # Add packet to the heap if we have not visited the node with
                                #   this packet before
                                priority = self.shortest_path_len(adj_dpid, dst_dpid)
                                dfs.push(adj_dpid, new_pkt, priority)
                                dfs.visit(adj_dpid, new_pkt)
                        elif trace:
                            # Output to non-stack port, can ignore this output
                            sys.stderr.write(
                                "Ignoring non-stack output %s:%s\n"
                                % (valve.dp.name, out_port)
                            )
            if trace:
                sys.stderr.write("\n")
        return found

    def table_state(self, dp_id):
        """Return tuple of table hash & table str"""
        return self.tables[dp_id].table_state()

    def hash_table(self, dp_id):
        """Return a hash of a single FakeOFTable"""
        return self.tables[dp_id].__hash__()


class FakeOFTable:
    """Fake OFTable is a virtual openflow pipeline used for testing openflow
    controllers.

    The tables are populated using apply_ofmsgs and can be queried with
    is_output.
    """

    def __init__(self, dp_id, num_tables=1, requires_tfm=True):
        self.dp_id = dp_id
        self.tables = [[] for _ in range(0, num_tables)]
        self.groups = {}
        self.requires_tfm = requires_tfm
        self.tfm = {}

    def table_state(self):
        """Return tuple of table hash & table str"""
        table_str = str(self.tables)
        return (hash(frozenset(table_str)), table_str)

    def __hash__(self):
        """Return a host of the tables"""
        return hash(frozenset(str(self.tables)))

    def _apply_groupmod(self, ofmsg):
        """Maintain group table."""

        def _del(_ofmsg, group_id):
            if group_id == ofp.OFPG_ALL:
                self.groups = {}
                return
            if group_id in self.groups:
                del self.groups[group_id]

        def _add(ofmsg, group_id):
            if group_id in self.groups:
                raise FakeOFTableException("group already in group table: %s" % ofmsg)
            self.groups[group_id] = ofmsg

        def _modify(ofmsg, group_id):
            if group_id not in self.groups:
                raise FakeOFTableException("group not in group table: %s" % ofmsg)
            self.groups[group_id] = ofmsg

        _groupmod_handlers = {
            ofp.OFPGC_DELETE: _del,
            ofp.OFPGC_ADD: _add,
            ofp.OFPGC_MODIFY: _modify,
        }

        _groupmod_handlers[ofmsg.command](ofmsg, ofmsg.group_id)

    def _apply_flowmod(self, ofmsg):
        """Adds, Deletes and modify flow modification messages are applied
        according to section 6.4 of the OpenFlow 1.3 specification."""

        def _validate_flowmod_tfm(table_id, tfm_body, ofmsg):
            if not self.requires_tfm:
                return

            if table_id == ofp.OFPTT_ALL:
                if ofmsg.match.items() and not self.tfm:
                    raise FakeOFTableException(
                        "got %s with matches before TFM that defines tables" % ofmsg
                    )
                return

            if tfm_body is None:
                raise FakeOFTableException(
                    "got %s before TFM that defines table %u" % (ofmsg, table_id)
                )

        def _add(table, flowmod):
            # From the 1.3 spec, section 6.4:
            # For add requests (OFPFC_ADD) with the
            # OFPFF_CHECK_OVERLAP flag set, the switch must first
            # check for any overlapping flow entries in the
            # requested table.  Two flow entries overlap if a
            # single packet may match both, and both flow entries
            # have the same priority, but the two flow entries
            # don't have the exact same match.  If an overlap
            # conflict exists between an existing flow entry and
            # the add request, the switch must refuse the addition
            # and respond with an ofp_error_msg with
            # OFPET_FLOW_MOD_FAILED type and OFPFMFC_OVERLAP code.
            #
            # Without the check overlap flag it seems like it is
            # possible that we can have overlapping flow table
            # entries which will cause ambiguous behaviour. This is
            # obviously unnacceptable so we will assume this is
            # always set
            for fte in table:
                if flowmod.fte_matches(fte, strict=True):
                    table.remove(fte)
                    break
                if flowmod.overlaps(fte):
                    raise FakeOFTableException(
                        "Overlapping flowmods {} and {}".format(flowmod, fte)
                    )
            table.append(flowmod)

        def _del(table, flowmod):
            removals = [fte for fte in table if flowmod.fte_matches(fte)]
            for fte in removals:
                table.remove(fte)

        def _del_strict(table, flowmod):
            for fte in table:
                if flowmod.fte_matches(fte, strict=True):
                    table.remove(fte)
                    break

        def _modify(table, flowmod):
            for fte in table:
                if flowmod.fte_matches(fte):
                    fte.instructions = flowmod.instructions

        def _modify_strict(table, flowmod):
            for fte in table:
                if flowmod.fte_matches(fte, strict=True):
                    fte.instructions = flowmod.instructions
                    break

        _flowmod_handlers = {
            ofp.OFPFC_ADD: _add,
            ofp.OFPFC_DELETE: _del,
            ofp.OFPFC_DELETE_STRICT: _del_strict,
            ofp.OFPFC_MODIFY: _modify,
            ofp.OFPFC_MODIFY_STRICT: _modify_strict,
        }

        table_id = ofmsg.table_id
        tfm_body = self.tfm.get(table_id, None)

        if table_id == ofp.OFPTT_ALL or table_id is None:
            tables = self.tables
        else:
            tables = [self.tables[table_id]]

        _validate_flowmod_tfm(table_id, tfm_body, ofmsg)
        flowmod = FlowMod(ofmsg)

        for table in tables:
            _flowmod_handlers[ofmsg.command](table, flowmod)

        if tfm_body:
            for table in tables:
                entries = len(table)
                if entries > tfm_body.max_entries:
                    tfm_table_details = "%s : table %u %s full (%u/%u)" % (
                        self.dp_id,
                        table_id,
                        tfm_body.name,
                        entries,
                        tfm_body.max_entries,
                    )
                    flow_dump = "\n\n".join(
                        (tfm_table_details, str(ofmsg), str(tfm_body))
                    )
                    raise FakeOFTableException(flow_dump)

    def _apply_tfm(self, ofmsg):
        self.tfm = {body.table_id: body for body in ofmsg.body}

    def _apply_flowstats(self, ofmsg):
        """Update state of flow tables to match an OFPFlowStatsReply message.

        This assumes a tfm is not required."""
        self.tables = []
        self.requires_tfm = False
        self.tfm = {}
        for stat in ofmsg.body:
            while len(self.tables) <= stat.table_id:
                self.tables.append([])
            self.tables[stat.table_id].append(FlowMod(stat))

    def apply_ofmsgs(self, ofmsgs, ignore_errors=False):
        """Update state of test flow tables."""
        for ofmsg in ofmsgs:
            try:
                if isinstance(ofmsg, parser.OFPBarrierRequest):
                    continue
                if isinstance(ofmsg, parser.OFPPacketOut):
                    continue
                if isinstance(ofmsg, parser.OFPSetConfig):
                    continue
                if isinstance(ofmsg, parser.OFPSetAsync):
                    continue
                if isinstance(ofmsg, parser.OFPDescStatsRequest):
                    continue
                if isinstance(ofmsg, parser.OFPMeterMod):
                    # TODO: handle OFPMeterMod
                    continue
                if isinstance(ofmsg, parser.OFPTableFeaturesStatsRequest):
                    self._apply_tfm(ofmsg)
                    continue
                if isinstance(ofmsg, parser.OFPGroupMod):
                    self._apply_groupmod(ofmsg)
                    continue
                if isinstance(ofmsg, parser.OFPFlowMod):
                    self._apply_flowmod(ofmsg)
                    self.sort_tables()
                    continue
                if isinstance(ofmsg, parser.OFPFlowStatsReply):
                    self._apply_flowstats(ofmsg)
                    self.sort_tables()
                    continue
            except FakeOFTableException:
                if not ignore_errors:
                    raise
            if not ignore_errors:
                raise FakeOFTableException("Unsupported flow %s" % str(ofmsg))

    def single_table_lookup(self, match, table_id, trace=False):
        """
        Searches through a single table with `table_id` for entries
        that will be applied to the packet with fields represented by match
        Args:
            match (dict): A dictionary keyed by header field names with values
            table_id (int): The table ID to send the match packet through
            trace (bool): Print the trace of traversing the table
        Returns:
            matching_fte: First matching flowmod in the table
        """
        packet_dict = match.copy()
        table = self.tables[table_id]
        matching_fte = None
        # Find matching flowmods
        for fte in table:
            if fte.pkt_matches(packet_dict):
                matching_fte = fte
                break
        if trace:
            sys.stderr.write("%s: %s\n" % (table_id, matching_fte))
        return matching_fte

    def _process_instruction(self, match, instruction):
        """
        Process an instructions actions into an output dictionary
        Args:
            match (dict): A dictionary keyed by header field names with values
            instruction: The instruction being applied to the packet match
        Returns:
            outputs: OrderedDict of an output port to list of output packets
            packet_dict: final dictionary of the packet
        """
        outputs = OrderedDict()
        packet_dict = match.copy()
        pending_actions = []
        for action in instruction.actions:
            if action.type == ofp.OFPAT_OUTPUT:
                # Save the packet that is output to a port
                outputs.setdefault(action.port, [])
                outputs[action.port].append(packet_dict.copy())
                pending_actions = []
                continue

            pending_actions.append(action)

            if action.type == ofp.OFPAT_SET_FIELD:
                # Set field, modify a packet header
                packet_dict[action.key] = action.value
            elif action.type == ofp.OFPAT_PUSH_VLAN:
                if (
                    "vlan_vid" in packet_dict
                    and packet_dict["vlan_vid"] & ofp.OFPVID_PRESENT
                ):
                    # Pushing on another tag, so create another
                    #   field for the encapsulated VID
                    packet_dict["encap_vid"] = packet_dict["vlan_vid"]
                # Push the VLAN header to the packet
                packet_dict["vlan_vid"] = ofp.OFPVID_PRESENT
            elif action.type == ofp.OFPAT_POP_VLAN:
                # Remove VLAN header from the packet
                packet_dict.pop("vlan_vid")
                if "vlan_pcp" in packet_dict:
                    # Also make sure to pop off any VLAN header information too
                    packet_dict.pop("vlan_pcp")
                if "encap_vid" in packet_dict:
                    # Move the encapsulated VID to the front
                    packet_dict["vlan_vid"] = packet_dict["encap_vid"]
                    packet_dict.pop("encap_vid")
                else:
                    packet_dict["vlan_vid"] = 0
            elif action.type == ofp.OFPAT_GROUP:
                # Group mod so make sure that we process the group buckets
                if action.group_id not in self.groups:
                    raise FakeOFTableException(
                        "output group not in group table: %s" % action
                    )
                buckets = self.groups[action.group_id].buckets
                for bucket in buckets:
                    bucket_outputs, _, _ = self._process_instruction(
                        packet_dict, bucket
                    )
                    for out_port, out_pkts in bucket_outputs.items():
                        outputs.setdefault(out_port, [])
                        outputs[out_port].extend(out_pkts)
                        pending_actions = []

        return outputs, packet_dict, pending_actions

    def get_table_output(self, match, table_id, trace=False):
        """
        Send a packet through a single table and return the output
        ports mapped to the output packet
        Args:
            match (dict): A dictionary keyed by header field names with values
            table_id (int): The table ID to send the packet match through
            trace (bool): Print the trace of traversing the table
        Returns:
            outputs: OrderedDict of an output port to output packet map
            packet_dict: The last version of the packet
            next_table: Table ID of the next table
        """
        next_table = None
        packet_dict = match.copy()
        outputs = OrderedDict()
        matching_fte = self.single_table_lookup(match, table_id, trace)
        pending_actions = []
        if matching_fte:
            for instruction in matching_fte.instructions:
                if instruction.type == ofp.OFPIT_GOTO_TABLE:
                    if table_id < instruction.table_id:
                        next_table = instruction.table_id
                    else:
                        raise FakeOFTableException("goto to lower table ID")
                elif instruction.type == ofp.OFPIT_APPLY_ACTIONS:
                    if not instruction.actions:
                        raise FakeOFTableException("no-op instruction actions")
                    (
                        instruction_outputs,
                        packet_dict,
                        pending_actions,
                    ) = self._process_instruction(packet_dict, instruction)
                    for out_port, out_pkts in instruction_outputs.items():
                        outputs.setdefault(out_port, [])
                        outputs[out_port].extend(out_pkts)
                elif instruction.type == ofp.OFPIT_WRITE_METADATA:
                    metadata = packet_dict.get("metadata", 0)
                    mask = instruction.metadata_mask
                    mask_compl = mask ^ 0xFFFFFFFFFFFFFFFF
                    packet_dict["metadata"] = (metadata & mask_compl) | (
                        instruction.metadata & mask
                    )
        if next_table:
            pending_actions = []
        if pending_actions:
            raise FakeOFTableException(
                "flow performs actions on packet after \
                                       output with no goto: %s"
                % matching_fte
            )
        return outputs, packet_dict, next_table

    def get_output(self, match, trace=False):
        """
        Get all of the outputs of the tables with the output packets
        for each table in the FakeOFTable that match progresses through
        Args:
            match (dict): A dictionary keyed by header field names with values
            trace (bool): Print the trace of traversing the table
        Returns:
            table_outputs: map from table_id output to output ports & packets
                for that table
        """
        table_outputs = {}
        table_id = 0
        next_table = True
        packet_dict = match.copy()
        while next_table:
            next_table = False
            outputs, packet_dict, next_table_id = self.get_table_output(
                packet_dict, table_id, trace
            )
            table_outputs[table_id] = outputs
            next_table = next_table_id is not None
            table_id = next_table_id
        return table_outputs

    def get_port_outputs(self, match, trace=False):
        """
        Get all of the outputs of the tables with the output packets
        for each table in the FakeOFTable that match progresses through
        Args:
            match (dict): A dictionary keyed by header field names with value
            trace (bool): Print the trace of traversing the table
        Returns:
            table_outputs: Map from output port number to a list of unique output packets
        """
        port_outputs = {}
        table_id = 0
        next_table = True
        packet_dict = match.copy()
        while next_table:
            next_table = False
            outputs, packet_dict, next_table_id = self.get_table_output(
                packet_dict, table_id, trace
            )
            for out_port, out_pkts in outputs.items():
                port_outputs.setdefault(out_port, [])
                # Remove duplicate entries from the list
                for out_pkt in out_pkts:
                    if out_pkt not in port_outputs[out_port]:
                        port_outputs[out_port].append(out_pkt)
            next_table = next_table_id is not None
            table_id = next_table_id
        return port_outputs

    def is_full_output(self, match, port=None, vid=None, trace=False):
        """
        If port is None return True if output to any port (including special ports)
        regardless of VLAN tag.
        If vid is None return True if output to specified port regardless of VLAN tag.
        If vid OFPVID_PRESENT bit is 0, return True if output packet does not have
        a VLAN tag OR packet OFPVID_PRESENT is 0
        Args:
            match (dict): A dictionary keyed by header field names with values
            port: The expected output port
            vid: The expected output vid
            trace (bool): Print the trace of traversing the tables
        Returns:
            true if packets with match fields is output to port with correct VLAN
        """
        table_outputs = self.get_output(match, trace)
        if trace:
            sys.stderr.write(pprint.pformat(table_outputs) + "\n")
        in_port = match.get("in_port")
        for table_outputs in table_outputs.values():
            for out_port, out_pkts in table_outputs.items():
                for out_pkt in out_pkts:
                    if port == out_port and port == out_pkt["in_port"]:
                        continue
                    if port is None:
                        # Port is None & outputting so return true
                        return True
                    if vid is None:
                        # Vid is None, return true if output to specified port
                        if port == out_port:
                            return True
                        if out_port == ofp.OFPP_IN_PORT and port == in_port:
                            # In some cases we want to match to specifically ofp.OFPP_IN_PORT
                            #   otherwise we treat ofp.OFPP_IN_PORT as the match in_port
                            return True
                    if port == out_port or (
                        out_port == ofp.OFPP_IN_PORT and port == in_port
                    ):
                        # Matching port, so check matching VID
                        if vid & ofp.OFPVID_PRESENT == 0:
                            # If OFPVID_PRESENT bit is 0 then packet should not have a VLAN tag
                            return (
                                "vlan_vid" not in out_pkt
                                or out_pkt["vlan_vid"] & ofp.OFPVID_PRESENT == 0
                            )
                        # VID specified, check if matching expected
                        return "vlan_vid" in out_pkt and vid == out_pkt["vlan_vid"]
        return False

    def lookup(self, match, trace=False):
        """Return the entries from flowmods that matches match.

        Searches each table in the pipeline for the entries that will be
        applied to the packet with fields represented by match.

        Arguments:
        match: a dictionary keyed by header field names with values.
                header fields not provided in match must be wildcarded for the
                entry to be considered matching.

        Returns: a list of the flowmods that will be applied to the packet
                represented by match
        """
        packet_dict = match.copy()  # Packet headers may be modified
        instructions = []
        table_id = 0
        goto_table = True
        while goto_table:
            goto_table = False
            table = self.tables[table_id]
            matching_fte = None
            # find a matching flowmod
            for fte in table:
                if fte.pkt_matches(packet_dict):
                    matching_fte = fte
                    break
            # if a flowmod is found, make modifications to the match values and
            # determine if another lookup is necessary
            if trace:
                sys.stderr.write("%d: %s\n" % (table_id, matching_fte))
            if matching_fte:
                for instruction in matching_fte.instructions:
                    instructions.append(instruction)
                    if instruction.type == ofp.OFPIT_GOTO_TABLE:
                        if table_id < instruction.table_id:
                            table_id = instruction.table_id
                            goto_table = True
                    elif instruction.type == ofp.OFPIT_APPLY_ACTIONS:
                        for action in instruction.actions:
                            if action.type == ofp.OFPAT_SET_FIELD:
                                packet_dict[action.key] = action.value
                    elif instruction.type == ofp.OFPIT_WRITE_METADATA:
                        metadata = packet_dict.get("metadata", 0)
                        mask = instruction.metadata_mask
                        mask_compl = mask ^ 0xFFFFFFFFFFFFFFFF
                        packet_dict["metadata"] = (metadata & mask_compl) | (
                            instruction.metadata & mask
                        )
        return (instructions, packet_dict)

    def flow_count(self):
        """Return number of flow tables rules"""
        return sum(map(len, self.tables))

    def is_output(self, match, port=None, vid=None, trace=False):
        """Return true if packets with match fields is output to port with
        correct vlan.

        If port is none it will return true if output to any port (including
        special ports) regardless of vlan tag.

        If vid is none it will return true if output to specified port
        regardless of vlan tag.

        To specify checking that the packet should not have a vlan tag, set the
        OFPVID_PRESENT bit in vid to 0.

        Arguments:
        Match: a dictionary keyed by header field names with values.
        """

        full_output = self.is_full_output(match.copy(), port, vid, trace)

        def _output_result(action, vid_stack, port, vid):
            if port is None:
                return True
            in_port = match.get("in_port")
            result = None
            if action.port == port:
                if port == in_port:
                    result = None
                elif vid is None:
                    result = True
                elif vid & ofp.OFPVID_PRESENT == 0:
                    result = not vid_stack
                else:
                    result = bool(vid_stack and vid == vid_stack[-1])
            elif action.port == ofp.OFPP_IN_PORT and port == in_port:
                result = True
            return result

        def _process_vid_stack(action, vid_stack):
            if action.type == ofp.OFPAT_PUSH_VLAN:
                vid_stack.append(ofp.OFPVID_PRESENT)
            elif action.type == ofp.OFPAT_POP_VLAN:
                vid_stack.pop()
            elif action.type == ofp.OFPAT_SET_FIELD:
                if action.key == "vlan_vid":
                    vid_stack[-1] = action.value
            return vid_stack

        if trace:
            sys.stderr.write(
                "tracing packet flow %s matching to port %s, vid %s\n"
                % (match, port, vid)
            )

        # vid_stack represents the packet's vlan stack, innermost label listed
        # first
        match_vid = match.get("vlan_vid", 0)
        vid_stack = []
        if match_vid & ofp.OFPVID_PRESENT != 0:
            vid_stack.append(match_vid)
        instructions, _ = self.lookup(match, trace=trace)

        for instruction in instructions:
            if instruction.type != ofp.OFPIT_APPLY_ACTIONS:
                continue
            for action in instruction.actions:
                vid_stack = _process_vid_stack(action, vid_stack)
                if action.type == ofp.OFPAT_OUTPUT:
                    output_result = _output_result(action, vid_stack, port, vid)
                    if output_result is not None:
                        if output_result != full_output:
                            raise FakeOFTableException("Output functions do not match")
                        return output_result
                elif action.type == ofp.OFPAT_GROUP:
                    if action.group_id not in self.groups:
                        raise FakeOFTableException(
                            "output group not in group table: %s" % action
                        )
                    buckets = self.groups[action.group_id].buckets
                    for bucket in buckets:
                        bucket_vid_stack = vid_stack
                        for bucket_action in bucket.actions:
                            bucket_vid_stack = _process_vid_stack(
                                bucket_action, bucket_vid_stack
                            )
                            if bucket_action.type == ofp.OFPAT_OUTPUT:
                                output_result = _output_result(
                                    bucket_action, vid_stack, port, vid
                                )
                                if output_result is not None:
                                    if output_result != full_output:
                                        raise FakeOFTableException(
                                            "Output functions do not match"
                                        )
                                    return output_result
        if full_output is not False:
            raise FakeOFTableException("Output functions do not match")
        return False

    def apply_instructions_to_packet(self, match):
        """
        Send packet through the fake OF table pipeline
        Args:
            match (dict): A dict keyed by header fields with values, represents
                a packet
        Returns:
            dict: Modified match dict, represents packet that has been through
                the pipeline with values possibly altered
        """
        _, packet_dict = self.lookup(match)
        return packet_dict

    def __str__(self):
        string = ""
        for table_id, table in enumerate(self.tables):
            string += "\n----- Table %u -----\n" % (table_id)
            string += "\n".join(sorted([str(flowmod) for flowmod in table]))
        return string

    def sort_tables(self):
        """Sort flows in tables by priority order."""
        self.tables = [sorted(table, reverse=True) for table in self.tables]


class FlowMod:
    """Represents a flow modification message and its corresponding entry in
    the flow table.
    """

    MAC_MATCH_FIELDS = (
        "eth_src",
        "eth_dst",
        "arp_sha",
        "arp_tha",
        "ipv6_nd_sll",
        "ipv6_nd_tll",
    )
    IPV4_MATCH_FIELDS = ("ipv4_src", "ipv4_dst", "arp_spa", "arp_tpa")
    IPV6_MATCH_FIELDS = ("ipv6_src", "ipv6_dst", "ipv6_nd_target")
    HEX_FIELDS = "eth_type"

    def __init__(self, flowmod):
        """flowmod is a ryu flow modification message object"""
        self.priority = flowmod.priority
        self.cookie = flowmod.cookie
        self.instructions = flowmod.instructions
        self.validate_instructions()
        self.match_values = {}
        self.match_masks = {}
        self.out_port = None
        # flowmod can be an OFPFlowMod or an OFPStats
        if isinstance(flowmod, parser.OFPFlowMod):
            if (
                flowmod.command in (ofp.OFPFC_DELETE, ofp.OFPFC_DELETE_STRICT)
                and flowmod.out_port != ofp.OFPP_ANY
            ):
                self.out_port = flowmod.out_port

        for key, val in flowmod.match.items():
            if isinstance(val, tuple):
                val, mask = val
            else:
                mask = -1

            mask = self.match_to_bits(key, mask)
            val = self.match_to_bits(key, val) & mask
            self.match_values[key] = val
            self.match_masks[key] = mask

    def validate_instructions(self):
        instruction_types = set()
        for instruction in self.instructions:
            if instruction.type in instruction_types:
                raise FakeOFTableException(
                    "FlowMod with Multiple instructions of the "
                    "same type: {}".format(self.instructions)
                )
            instruction_types.add(instruction.type)

    def out_port_matches(self, other):
        """returns True if other has an output action to this flowmods
        output_port"""
        if self.out_port is None or self.out_port == ofp.OFPP_ANY:
            return True
        for instruction in other.instructions:
            if instruction.type == ofp.OFPIT_APPLY_ACTIONS:
                for action in instruction.actions:
                    if action.type == ofp.OFPAT_OUTPUT:
                        if action.port == self.out_port:
                            return True
        return False

    def pkt_matches(self, pkt_dict):
        """returns True if pkt_dict matches this flow table entry.

        args:
            pkt_dict - a dictionary keyed by flow table match fields with
                values
        if an element is included in the flow table entry match fields but not
        in the pkt_dict that is assumed to indicate a failed match
        """

        # TODO: add cookie and out_group
        for key, val in self.match_values.items():
            if key not in pkt_dict:
                return False
            val_bits = self.match_to_bits(key, pkt_dict[key])
            if val_bits != (val & self.match_masks[key]):
                return False
        return True

    def _matches_match(self, other):
        return (
            self.priority == other.priority
            and self.match_values == other.match_values
            and self.match_masks == other.match_masks
        )

    def fte_matches(self, other, strict=False):
        """returns True if the flow table entry other matches this flowmod.

        used for finding existing flow table entries that match with this
        flowmod.

        args:
            other - a flowmod object
            strict (bool) - whether to use strict matching (as defined in
                of1.3 specification section 6.4)
        """
        if not self.out_port_matches(other):
            return False
        if strict:
            return self._matches_match(other)
        for key, val in self.match_values.items():
            if key not in other.match_values:
                return False
            if other.match_values[key] & self.match_masks[key] != val:
                return False
        return True

    def overlaps(self, other):
        """returns True if any packet can match both self and other."""
        # This is different from the matches method as matches assumes an
        # undefined field is a failed match. In this case an undefined field is
        # potentially an overlap and therefore is considered success
        if other.priority != self.priority:
            return False
        for key, val in self.match_values.items():
            if key in other.match_values:
                if val & other.match_masks[key] != other.match_values[key]:
                    return False
                if other.match_values[key] & self.match_masks[key] != val:
                    return False
        return True

    def match_to_bits(self, key, val):
        """convert match fields and masks to bits objects.

        this allows for masked matching. Converting all match fields to the
        same object simplifies things (eg __str__).
        """
        if isinstance(val, Bits):
            return val

        def _val_to_bits(conv, val, length):
            if val == -1:
                return Bits(int=-1, length=length)
            return Bits(bytes=conv(val), length=length)

        if key in self.MAC_MATCH_FIELDS:
            return _val_to_bits(addrconv.mac.text_to_bin, val, 48)
        if key in self.IPV4_MATCH_FIELDS:
            return _val_to_bits(addrconv.ipv4.text_to_bin, val, 32)
        if key in self.IPV6_MATCH_FIELDS:
            return _val_to_bits(addrconv.ipv6.text_to_bin, val, 128)
        return Bits(int=int(val), length=64)

    def bits_to_str(self, key, val):
        if key in self.MAC_MATCH_FIELDS:
            result = addrconv.mac.bin_to_text(val.tobytes())
        elif key in self.IPV4_MATCH_FIELDS:
            result = addrconv.ipv4.bin_to_text(val.tobytes())
        elif key in self.IPV6_MATCH_FIELDS:
            result = addrconv.ipv6.bin_to_text(val.tobytes())
        elif key in self.HEX_FIELDS:
            result = str(val.hex.lstrip("0"))
        else:
            result = str(val.int)
        return result

    def __lt__(self, other):
        return self.priority < other.priority

    def __eq__(self, other):
        return (
            self._matches_match(other)
            and self.out_port == other.out_port
            and self.instructions == other.instructions
        )

    def __hash__(self):
        return hash(
            (
                self.priority,
                self.match_values,
                self.match_masks,
                self.out_port,
                self.instructions,
            )
        )

    def _pretty_field_str(self, key, value, mask=None):
        mask_str = ""
        value_int = value
        mask_int = mask
        if isinstance(value, Bits):
            value_int = value.int
        if isinstance(mask, Bits):
            mask_int = mask.int  # pytype: disable=attribute-error
        elif mask is None:
            mask_int = -1
        if key == "vlan_vid":
            if value_int & ofp.OFPVID_PRESENT == 0:
                result = "vlan untagged"
            elif key == "vlan_vid" and mask_int == ofp.OFPVID_PRESENT:
                result = "vlan tagged"
            else:
                result = str(value_int ^ ofp.OFPVID_PRESENT)
                if mask_int != -1:
                    mask_str = str(mask_int ^ ofp.OFPVID_PRESENT)
        elif isinstance(value, Bits):
            result = self.bits_to_str(key, value)
            if mask is not None and mask_int != -1:
                mask_str = self.bits_to_str(key, mask)
        elif isinstance(value, str):
            result = value
            if mask is not None:
                mask_str = mask
        elif isinstance(value, int):
            if key in self.HEX_FIELDS:
                result = hex(value)
                if mask is not None and mask != -1:
                    mask_str = hex(mask)
            else:
                result = str(value)
                if mask is not None and mask != -1:
                    mask_str = str(mask)
        if mask_str:
            result += "/{}".format(mask_str)
        return result

    def _pretty_action_str(self, action):
        actions_names_attrs = {
            parser.OFPActionPushVlan.__name__: ("push_vlan", "ethertype"),
            parser.OFPActionPopVlan.__name__: ("pop_vlan", None),
            parser.OFPActionGroup.__name__: ("group", "group_id"),
            parser.OFPActionDecNwTtl.__name__: ("dec_nw_ttl", None),
        }
        value = None
        if isinstance(action, parser.OFPActionOutput):
            name = "output"
            if action.port == CONTROLLER_PORT:
                value = "CONTROLLER"
            elif action.port == IN_PORT:
                value = "IN_PORT"
            else:
                value = str(action.port)
        elif isinstance(action, parser.OFPActionSetField):
            name = "set_{}".format(action.key)
            value = self._pretty_field_str(action.key, action.value)
        else:
            name, attr = actions_names_attrs[type(action).__name__]
            if attr:
                value = getattr(action, attr)
        result = name
        if value:
            result += " {}".format(value)
        return result

    def __str__(self):
        result = "Priority: {0} | Match: ".format(self.priority)

        for key in sorted(self.match_values.keys()):
            val = self.match_values[key]
            mask = self.match_masks[key]
            result += " {} {},".format(key, self._pretty_field_str(key, val, mask))
        result = result.rstrip(",")
        result += " | Instructions :"
        if not self.instructions:
            result += " drop"
        for instruction in self.instructions:
            if isinstance(instruction, parser.OFPInstructionGotoTable):
                result += " goto {}".format(instruction.table_id)
            elif isinstance(instruction, parser.OFPInstructionActions):
                for action in instruction.actions:
                    result += " {},".format(self._pretty_action_str(action))
            else:
                result += str(instruction)
        result = result.rstrip(",")
        return result

    def __repr__(self):
        string = "priority: {0} cookie: {1}".format(self.priority, self.cookie)
        for key in sorted(self.match_values.keys()):
            mask = self.match_masks[key]
            string += " {0}: {1}".format(key, self.match_values[key])
            if mask.int != -1:  # pytype: disable=attribute-error
                string += "/{0}".format(mask)
        string += " Instructions: {0}".format(str(self.instructions))
        return string


class FakeRyuDp:  # pylint: disable=too-few-public-methods
    """Fake ryu Datapath object.

    Just needed to provide a parser to allow us to extract ryu objects from
    JSON
    """

    def __init__(self):
        """Create fake ryu DP"""
        self.ofproto_parser = parser


def parse_print_args():
    """Parse arguments for the print command"""
    arg_parser = argparse.ArgumentParser(
        prog="fakeoftable",
        description="Prints a JSON flow table in a human readable format",
        usage="""
    Print a flow table in a human readable format
    {argv0} print -f FILE
""".format(
            argv0=sys.argv[0]
        ),
    )
    arg_parser.add_argument(
        "-f",
        "--file",
        help="file containing an OFPFlowStatsReply message in JSON format",
    )
    args = arg_parser.parse_args(sys.argv[2:])
    return {"filename": args.file}


def parse_probe_args():
    """Parse arguments for the probe command"""
    arg_parser = argparse.ArgumentParser(
        prog="fakeoftable",
        description="Performs a packet lookup on a JSON openflow table",
        usage="""
    Find the flow table entries in a given flow table that match a given packet
    {argv0} probe -f FILE -p PACKET_STRING
""".format(
            argv0=sys.argv[0]
        ),
    )
    arg_parser.add_argument(
        "-p",
        "--packet",
        metavar="PACKET_STRING",
        help=(
            """string representation of a packet dictionary eg. """
            """"{'in_port': 1, 'eth_dst': '01:80:c2:00:00:02', 'eth_type': """
            '''34825}"'''
        ),
    )
    arg_parser.add_argument(
        "-f",
        "--file",
        metavar="FILE",
        help="file containing an OFPFlowStatsReply message in JSON format",
    )
    args = arg_parser.parse_args(sys.argv[2:])
    packet = args.packet
    packet = ast.literal_eval(args.packet)
    # fix vlan vid
    if "vlan_vid" in packet:
        packet["vlan_vid"] |= ofp.OFPVID_PRESENT
    return {"packet": packet, "filename": args.file}


def parse_args():
    """parse arguments"""
    arg_parser = argparse.ArgumentParser(
        prog="fakeoftable",
        description="Performs operations on JSON openflow tables",
        usage="""
    {argv0} <command> <args>

""".format(
            argv0=sys.argv[0]
        ),
    )
    arg_parser.add_argument("command", help='Subcommand, either "print" or "probe"')
    args = arg_parser.parse_args(sys.argv[1:2])
    try:
        if args.command == "probe":
            command_args = parse_probe_args()
        elif args.command == "print":
            command_args = parse_print_args()
    except (KeyError, IndexError, ValueError, AttributeError) as err:
        print(err)
        arg_parser.print_help()
        sys.exit(-1)
    return (args.command, command_args)


def _print(filename, **_kwargs):
    """Prints the JSON flow table from a file in a human readable format"""
    with open(filename, "r", encoding="utf-8") as file_handle:
        msg = json.load(file_handle)
    datapath = FakeRyuDp()
    ofmsg = ofp_parser.ofp_msg_from_jsondict(datapath, msg)
    table = FakeOFTable(1)
    table.apply_ofmsgs([ofmsg])
    print(table)


def probe(filename, packet):
    """Prints the actions applied to packet by the table from the file"""
    with open(filename, "r", encoding="utf-8") as file_handle:
        msg = json.load(file_handle)
    datapath = FakeRyuDp()
    ofmsg = ofp_parser.ofp_msg_from_jsondict(datapath, msg)
    table = FakeOFTable(1)
    table.apply_ofmsgs([ofmsg])
    instructions, out_packet = table.lookup(packet)
    print(packet)
    for instruction in instructions:
        print(instruction)
    print(out_packet)


def main():
    command, kwargs = parse_args()
    if command == "probe":
        probe(**kwargs)
    elif command == "print":
        _print(**kwargs)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Faucet setup script"""

#  pylint: disable=consider-using-f-string

from __future__ import print_function

import errno
import os
import shutil
import sys

from pkg_resources import resource_filename
from setuptools import setup

if sys.version_info < (3,):
    print(
        """You are trying to install faucet on python {py}

Faucet is not compatible with python 2, please upgrade to python 3.9 or newer.""".format(
            py=".".join([str(v) for v in sys.version_info[:3]])
        ),
        file=sys.stderr,
    )
    sys.exit(1)
elif sys.version_info < (3, 9):
    print(
        """You are trying to install faucet on python {py}

Faucet 1.9.0 and above are no longer compatible with older versions of python 3.

Please upgrade to python 3.9 or newer.""".format(
            py=".".join([str(v) for v in sys.version_info[:3]])
        )
    )
    sys.exit(1)


def install_configs():
    """Install configuration files to /etc"""

    dst_ryu_conf_dir = "/etc/faucet/"
    dst_ryu_conf = os.path.join(dst_ryu_conf_dir, "ryu.conf")
    dst_faucet_conf_dir = "/etc/faucet/"
    src_ryu_conf = resource_filename(__name__, "etc/faucet/ryu.conf")
    src_faucet_conf_dir = resource_filename(__name__, "etc/faucet/")
    faucet_log_dir = "/var/log/faucet/"

    old_ryu_conf = "/etc/ryu/ryu.conf"
    old_faucet_conf_dir = "/etc/ryu/faucet/"

    def setup_ryu_conf():
        if not os.path.exists(dst_ryu_conf_dir):
            print("Creating %s" % dst_ryu_conf_dir)
            os.makedirs(dst_ryu_conf_dir)
        if not os.path.isfile(dst_ryu_conf):
            if os.path.exists(old_ryu_conf) and os.path.isfile(old_ryu_conf):
                print("Migrating %s to %s" % (old_ryu_conf, dst_ryu_conf))
                shutil.copy(old_ryu_conf, dst_ryu_conf)
            else:
                print("Copying %s to %s" % (src_ryu_conf, dst_ryu_conf))
                shutil.copy(src_ryu_conf, dst_ryu_conf)

    def setup_faucet_conf():
        if not os.path.exists(dst_faucet_conf_dir):
            print("Creating %s" % dst_faucet_conf_dir)
            os.makedirs(dst_faucet_conf_dir)
        for file_name in os.listdir(src_faucet_conf_dir):
            src_file = os.path.join(src_faucet_conf_dir, file_name)
            dst_file = os.path.join(dst_faucet_conf_dir, file_name)
            alt_src = os.path.join(old_faucet_conf_dir, file_name)
            if os.path.isfile(dst_file):
                continue
            if os.path.isfile(alt_src):
                print("Migrating %s to %s" % (alt_src, dst_file))
                shutil.copy(alt_src, dst_file)
            elif os.path.isfile(src_file):
                print("Copying %s to %s" % (src_file, dst_file))
                shutil.copy(src_file, dst_file)

    def setup_faucet_log():
        if not os.path.exists(faucet_log_dir):
            print("Creating %s" % faucet_log_dir)
            os.makedirs(faucet_log_dir)

    try:
        setup_ryu_conf()
        setup_faucet_conf()
        setup_faucet_log()
    except OSError as exception:
        if exception.errno == errno.EACCES:
            print(
                "Permission denied creating %s, skipping copying configs"
                % exception.filename
            )
        elif exception.errno == errno.ENOENT:
            print(
                "File not found creating %s, skipping copying configs"
                % exception.filename
            )
        else:
            raise


setup(
    name="faucet",
    setup_requires=["pbr>=1.9", "setuptools>=17.1"],
    python_requires=">=3.9",
    pbr=True,
)

if "install" in sys.argv or "bdist_wheel" in sys.argv:
    if os.getenv("DEBINSTALL") is None or (
        os.getenv("DEBINSTALL") is not None and int(os.environ["DEBINSTALL"]) < 1
    ):
        install_configs()



#!/usr/bin/env python3

"""Standalone utility functions for Mininet tests."""

import collections
import os
import socket
import subprocess
import time

# pylint: disable=import-error
from mininet.log import error, output


DEVNULL = open(os.devnull, "wb", encoding=None)  # pylint: disable=consider-using-with
GETPORT = "GETPORT"
PUTPORTS = "PUTPORTS"
GETSERIAL = "GETSERIAL"
LISTPORTS = "LISTPORTS"
LOCALHOST = "127.0.0.1"
LOCALHOSTV6 = "::1"
FAUCET_DIR = os.getenv("FAUCET_DIR", "../faucet")
RESERVED_FOR_TESTS_PORTS = (179, 5001, 5002, 6633, 6653)
with open(
    "/proc/sys/net/netfilter/nf_conntrack_tcp_timeout_time_wait", encoding="utf-8"
) as pf:
    MIN_PORT_AGE = max(int(pf.read()) / 2, 10)


def flat_test_name(_id):
    """Return short form test name from TestCase ID."""
    return "-".join(_id.split(".")[1:])


def lsof_tcp_listening_cmd(port, ipv, state, terse, pid):
    """Return a command line for lsof for processes with specified TCP state."""
    terse_arg = ""
    if terse:
        terse_arg = "-t"
    pid_arg = ""
    if pid:
        pid_arg = "-p %u" % pid
    return "lsof -b -P -n %s %s -sTCP:%s -i %u -a -i tcp:%u" % (
        pid_arg,
        terse_arg,
        state,
        ipv,
        port,
    )


def lsof_udp_listening_cmd(port, terse):
    """Return a command line for lsof for processes with specified TCP state."""
    terse_arg = ""
    if terse:
        terse_arg = "-t"
    return "lsof -b -P -n %s -i udp:%u -a" % (terse_arg, port)


def tcp_listening_cmd(port, ipv=4, state="LISTEN", terse=True, pid=None):
    """Call lsof_tcp_listening_cmd() with default args."""
    return lsof_tcp_listening_cmd(port, ipv, state, terse, pid)


def udp_listening_cmd(port, terse=True):
    """Call lsof_tcp_listening_cmd() with default args."""
    return lsof_udp_listening_cmd(port, terse)


def mininet_dpid(int_dpid):
    """Return stringified hex version, of int DPID for mininet."""
    return str("%x" % int(int_dpid))


def str_int_dpid(str_dpid):
    """Return stringified int version, of int or hex DPID from YAML."""
    str_dpid = str(str_dpid)
    if str_dpid.startswith("0x"):
        return str(int(str_dpid, 16))
    return str(int(str_dpid))


def receive_sock_line(sock):
    """Receive a \n terminated line from a socket."""
    buf = ""
    while buf.find("\n") <= -1:
        buf += sock.recv(2**10).decode()
    return buf.strip()


def tcp_listening(port):
    """Return True if any process listening on a port."""
    return (
        subprocess.call(
            tcp_listening_cmd(port).split(),
            stdin=DEVNULL,
            stdout=DEVNULL,
            stderr=DEVNULL,
            close_fds=True,
        )
        == 0
    )


def udp_listening(port):
    """Return True if any process listening on a port."""
    return (
        subprocess.call(
            udp_listening_cmd(port).split(),
            stdin=DEVNULL,
            stdout=DEVNULL,
            stderr=DEVNULL,
            close_fds=True,
        )
        == 0
    )


def test_server_request(ports_socket, name, command):
    assert name is not None
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.connect(ports_socket)
    sock.sendall(("%s,%s\n" % (command, name)).encode())
    output("%s %s\n" % (name, command))
    buf = receive_sock_line(sock)
    responses = [int(i) for i in buf.split("\n")]
    sock.close()
    if len(responses) == 1:
        responses = responses[0]
    output("%s %s: %u\n" % (name, command, responses))
    return responses


def get_serialno(ports_socket, name):
    """Retrieve serial number from test server."""
    return test_server_request(ports_socket, name, GETSERIAL)


def find_free_port(ports_socket, name):
    """Retrieve a free TCP port from test server."""
    request_name = "-".join((name, str(os.getpid())))
    while True:
        port = test_server_request(ports_socket, request_name, GETPORT)
        if not tcp_listening(port):
            return port
        error("port %u is busy, try another" % port)


def find_free_udp_port(ports_socket, name):
    request_name = "-".join((name, str(os.getpid())))
    while True:
        port = test_server_request(ports_socket, request_name, GETPORT)
        if not udp_listening(port):
            return port
        error("port %u is busy, try another" % port)


def return_free_ports(ports_socket, name):
    """Notify test server that all ports under name are released."""
    return test_server_request(ports_socket, name, PUTPORTS)


def serve_ports(ports_socket, start_free_ports, min_free_ports):
    """Implement a TCP server to dispense free TCP ports."""
    ports_q = collections.deque()
    free_ports = set()
    port_age = {}
    serialno = 0

    def get_port():
        while True:
            free_socket = socket.socket()
            free_socket.bind(("", 0))
            free_port = free_socket.getsockname()[1]
            free_socket.close()
            if free_port < 1024:
                continue
            if free_port in RESERVED_FOR_TESTS_PORTS:
                continue
            if free_port in free_ports:
                continue
            break
        free_ports.add(free_port)
        port_age[free_port] = time.time()
        return free_port

    def queue_free_ports(min_queue_size):
        while len(ports_q) < min_queue_size:
            port = get_port()
            ports_q.append(port)
            port_age[port] = time.time()

    queue_free_ports(start_free_ports)
    ports_by_name = collections.defaultdict(set)
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.bind(ports_socket)
    sock.listen(1)
    cold_start = True

    while True:
        connection, _ = sock.accept()
        command, name = receive_sock_line(connection).split(",")
        response = None
        if command == GETSERIAL:
            serialno += 1
            response = serialno
        elif command == PUTPORTS:
            ports_returned = 0
            for port in ports_by_name[name]:
                ports_returned += 1
                ports_q.append(port)
                port_age[port] = time.time()
            del ports_by_name[name]
            response = ports_returned
            if ports_returned:
                cold_start = False
        elif command == GETPORT:
            while True:
                port = ports_q.popleft()
                if time.time() - port_age[port] > MIN_PORT_AGE or cold_start:
                    break
                ports_q.append(port)
                time.sleep(1)
            ports_by_name[name].add(port)
            response = port
            queue_free_ports(min_free_ports)
        elif command == LISTPORTS:
            response = list(ports_by_name[name])
        if response is not None:
            response_str = ""
            if isinstance(response, int):
                response = [response]
            response_str = "".join(["%u\n" % i for i in response])
            connection.sendall(response_str.encode())  # pylint: disable=no-member
        connection.close()


def timeout_cmd(cmd, timeout):
    """Return a command line prefaced with a timeout wrappers and stdout/err unbuffered."""
    return "timeout -sKILL %us stdbuf -o0 -e0 %s" % (timeout, cmd)


def timeout_soft_cmd(cmd, timeout):
    """Same as timeout_cmd buf using SIGTERM on timeout."""
    return "timeout %us stdbuf -o0 -e0 %s" % (timeout, cmd)



"""Abstraction of an OF table."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import hashlib
import struct
from faucet import valve_of
from faucet.faucet_pipeline import ValveTableConfig


class ValveTable:  # pylint: disable=too-many-arguments,too-many-instance-attributes
    """Wrapper for an OpenFlow table."""

    def __init__(
        self,
        name,
        table_config,
        flow_cookie,
        notify_flow_removed=False,
        next_tables=None,
    ):
        self.name = name
        self.table_config = table_config
        self.table_id = self.table_config.table_id
        self.set_fields = self.table_config.set_fields
        self.exact_match = self.table_config.exact_match
        self.match_types = None
        self.metadata_match = self.table_config.metadata_match
        self.metadata_write = self.table_config.metadata_write
        if next_tables:
            self.next_tables = next_tables
        else:
            self.next_tables = []
        if self.table_config.match_types:
            self.match_types = {}
            for field, mask in self.table_config.match_types:
                self.match_types[field] = mask
        self.flow_cookie = flow_cookie
        self.notify_flow_removed = notify_flow_removed

    def goto(self, next_table):
        """Add goto next table instruction."""
        assert (
            next_table.name in self.table_config.next_tables
        ), "%s not configured as next table in %s" % (next_table.name, self.name)
        return valve_of.goto_table(next_table)

    def goto_this(self):
        return valve_of.goto_table(self)

    def goto_miss(self, next_table):
        """Add miss goto table instruction."""
        assert (
            next_table.name == self.table_config.miss_goto
        ), "%s not configured as miss table in %s" % (next_table.name, self.name)
        return valve_of.goto_table(next_table)

    @staticmethod
    def set_field(**kwds):
        """Return set field action."""
        # raise exception if unknown set field.
        valve_of.match_from_dict(kwds)
        return valve_of.set_field(**kwds)

    def set_external_forwarding_requested(self):
        """Set field for external forwarding requested."""
        return self.set_field(
            **{valve_of.EXTERNAL_FORWARDING_FIELD: valve_of.PCP_EXT_PORT_FLAG}
        )

    def set_no_external_forwarding_requested(self):
        """Set field for no external forwarding requested."""
        return self.set_field(
            **{valve_of.EXTERNAL_FORWARDING_FIELD: valve_of.PCP_NONEXT_PORT_FLAG}
        )

    def set_vlan_vid(self, vlan_vid):
        """Set VLAN VID with VID_PRESENT flag set.

        Args:
            vid (int): VLAN VID
        Returns:
            ryu.ofproto.ofproto_v1_3_parser.OFPActionSetField: set VID with VID_PRESENT.
        """
        return self.set_field(vlan_vid=valve_of.vid_present(vlan_vid))

    # TODO: verify actions
    @staticmethod
    @functools.lru_cache(maxsize=1024)
    def match(
        in_port=None,
        vlan=None,  # pylint: disable=too-many-arguments
        eth_type=None,
        eth_src=None,
        eth_dst=None,
        eth_dst_mask=None,
        icmpv6_type=None,
        nw_proto=None,
        nw_dst=None,
        metadata=None,
        metadata_mask=None,
        vlan_pcp=None,
        udp_src=None,
        udp_dst=None,
    ):
        """Compose an OpenFlow match rule."""
        match_dict = valve_of.build_match_dict(
            in_port,
            vlan,
            eth_type,
            eth_src,
            eth_dst,
            eth_dst_mask,
            icmpv6_type,
            nw_proto,
            nw_dst,
            metadata,
            metadata_mask,
            vlan_pcp,
            udp_src,
            udp_dst,
        )
        return valve_of.match(match_dict)

    def _verify_flowmod(self, flowmod):
        match_fields = flowmod.match.items()
        if valve_of.is_flowdel(flowmod):
            if self.table_id != valve_of.ofp.OFPTT_ALL:
                for match_type, match_field in match_fields:
                    assert match_type in self.match_types, "%s match in table %s" % (
                        match_type,
                        self.name,
                    )
        else:
            # TODO: ACL builder should not use ALL table.
            if self.table_id == valve_of.ofp.OFPTT_ALL:
                return
            assert not (
                flowmod.priority == 0 and match_fields
            ), "default flow cannot have matches on table %s: %s" % (self.name, flowmod)
            for match_type, match_field in match_fields:
                assert match_type in self.match_types, "%s match in table %s" % (
                    match_type,
                    self.name,
                )
                config_mask = self.match_types[match_type]
                flow_mask = isinstance(match_field, tuple)
                assert config_mask or (
                    not config_mask and not flow_mask
                ), "%s configured mask %s but flow mask %s in table %s (%s)" % (
                    match_type,
                    config_mask,
                    flow_mask,
                    self.name,
                    flowmod,
                )
                if self.exact_match and match_fields:
                    assert len(self.match_types) == len(match_fields), (
                        "exact match table %s matches %s do not match flow matches %s (%s)"
                        % (self.name, self.match_types, match_fields, flowmod)
                    )

    def _trim_actions(self, actions):
        new_actions = []
        pending_actions = []
        for action in actions:
            if action.type in (valve_of.ofp.OFPAT_GROUP, valve_of.ofp.OFPAT_OUTPUT):
                new_actions.extend(pending_actions)
                new_actions.append(action)
                pending_actions = []
            else:
                pending_actions.append(action)
        set_fields = {
            action.key for action in new_actions if valve_of.is_set_field(action)
        }
        if self.table_id != valve_of.ofp.OFPTT_ALL and set_fields:
            assert set_fields.issubset(
                self.set_fields
            ), "unexpected set fields %s configured %s in %s" % (
                set_fields,
                self.set_fields,
                self.name,
            )
        return new_actions

    @functools.lru_cache()
    def _trim_inst(self, inst):
        """Discard empty/actions on packets that are not output and not goto another table."""
        inst_types = {instruction.type for instruction in inst}
        if valve_of.ofp.OFPIT_APPLY_ACTIONS in inst_types:
            goto_present = valve_of.ofp.OFPIT_GOTO_TABLE in inst_types
            new_inst = []
            for instruction in inst:
                if instruction.type == valve_of.ofp.OFPIT_APPLY_ACTIONS:
                    recirc_present = any(
                        (
                            True
                            for action in instruction.actions
                            if valve_of.is_ct(action)
                            and hasattr(action, "recirc_table")
                        )
                    )
                    # If no goto present, this is the last set of actions that can take place
                    if not goto_present and not recirc_present:
                        instruction.actions = self._trim_actions(instruction.actions)
                    # Always drop an apply actions instruction with no actions.
                    if not instruction.actions:
                        continue
                new_inst.append(instruction)
            return tuple(new_inst)
        return inst

    def flowmod(
        self,
        match=None,
        priority=None,  # pylint: disable=too-many-arguments
        inst=None,
        command=valve_of.ofp.OFPFC_ADD,
        out_port=0,
        out_group=0,
        hard_timeout=0,
        idle_timeout=0,
        cookie=None,
    ):
        """Helper function to construct a flow mod message with cookie."""
        if priority is None:
            priority = 0  # self.dp.lowest_priority
        if not match:
            match = self.match()
        if inst is None:
            inst = ()
        if cookie is None:
            cookie = self.flow_cookie
        flags = 0
        if self.notify_flow_removed:
            flags = valve_of.ofp.OFPFF_SEND_FLOW_REM
        if inst:
            inst = self._trim_inst(inst)
        flowmod = valve_of.flowmod(
            cookie,
            command,
            self.table_id,
            priority,
            out_port,
            out_group,
            match,
            tuple(inst),
            hard_timeout,
            idle_timeout,
            flags,
        )
        self._verify_flowmod(flowmod)
        return flowmod

    def flowdel(
        self, match=None, priority=None, out_port=valve_of.ofp.OFPP_ANY, strict=False
    ):
        """Delete matching flows from a table."""
        command = valve_of.ofp.OFPFC_DELETE
        if strict:
            command = valve_of.ofp.OFPFC_DELETE_STRICT
        return self.flowmod(
            match=match,
            priority=priority,
            command=command,
            out_port=out_port,
            out_group=valve_of.ofp.OFPG_ANY,
        )

    def flowdrop(self, match=None, priority=None, hard_timeout=0):
        """Add drop matching flow to a table."""
        return self.flowmod(
            match=match, priority=priority, hard_timeout=hard_timeout, inst=()
        )

    def flowcontroller(self, match=None, priority=None, inst=None, max_len=96):
        """Add flow outputting to controller."""
        if inst is None:
            inst = ()
        return self.flowmod(
            match=match,
            priority=priority,
            inst=(valve_of.apply_actions((valve_of.output_controller(max_len),)),)
            + inst,
        )


class ValveGroupEntry:
    """Abstraction for a single OpenFlow group entry."""

    def __init__(self, table, group_id, buckets):
        self.table = table
        self.group_id = group_id
        self.update_buckets(buckets)

    def update_buckets(self, buckets):
        """Update entry with new buckets."""
        self.buckets = tuple(buckets)

    def add(self):
        """Return flows to add this entry to the group table."""
        ofmsgs = []
        ofmsgs.extend(valve_of.groupadd(group_id=self.group_id, buckets=self.buckets))
        self.table.entries[self.group_id] = self
        return ofmsgs

    def delete(self):
        """Return flow to delete an existing group entry."""
        if self.group_id in self.table.entries:
            del self.table.entries[self.group_id]
        return valve_of.groupdel(group_id=self.group_id)


class ValveGroupTable:
    """Wrap access to group table."""

    entries = None  # type: dict

    def __init__(self):
        """Constructs a new object"""
        self.entries = {}

    @staticmethod
    def group_id_from_str(key_str):
        """Return a group ID based on a string key."""
        # TODO: does not handle collisions
        digest = hashlib.sha256(key_str.encode("utf-8")).digest()
        return struct.unpack("<L", digest[:4])[0]

    def get_entry(self, group_id, buckets):
        """Update entry with group_id with buckets, and return the entry."""
        if group_id in self.entries:
            self.entries[group_id].update_buckets(buckets)
        else:
            self.entries[group_id] = ValveGroupEntry(self, group_id, buckets)
        return self.entries[group_id]

    def delete_all(self):
        """Delete all groups."""
        self.entries = {}
        return valve_of.groupdel()


wildcard_table = ValveTable(
    "all", ValveTableConfig("all", valve_of.ofp.OFPTT_ALL), flow_cookie=0
)
